
declare namespace Van {

    interface State<T> {
      val: T
      readonly oldVal: T
      readonly rawVal: T
    }
    
    // Defining readonly view of State<T> for covariance.
    // Basically we want StateView<string> to implement StateView<string | number>
    type StateView<T> = Readonly<State<T>>
    
    type Val<T> = State<T> | T
    
    type Primitive = string | number | boolean | bigint
    
    // deno-lint-ignore no-explicit-any
    type PropValue = Primitive | ((e: any) => void) | null
    
    type PropValueOrDerived = PropValue | StateView<PropValue> | (() => PropValue)
    
    type Props = Record<string, PropValueOrDerived> & { class?: PropValueOrDerived; is?: string }
    
    type PropsWithKnownKeys<ElementType> = Partial<{[K in keyof ElementType]: PropValueOrDerived}>
    
    type ValidChildDomValue = Primitive | Node | null | undefined
    
    type BindingFunc = ((dom?: Node) => ValidChildDomValue) | ((dom?: Element) => Element)
    
    type ChildDom = ValidChildDomValue | StateView<Primitive | null | undefined> | BindingFunc | readonly ChildDom[]
    
    type TagFunc<Result> = (first?: Props & PropsWithKnownKeys<Result> | ChildDom, ...rest: readonly ChildDom[]) => Result
    
    type Tags = Readonly<Record<string, TagFunc<Element>>> & {
      [K in keyof HTMLElementTagNameMap]: TagFunc<HTMLElementTagNameMap[K]>
    }
}
declare namespace VanX
{
    type StateOf<T> = { readonly [K in keyof T]: Van.State<T[K]> }
    type ValueType<T> = T extends (infer V)[] ? V : T[keyof T]
    type KeyType<T> = T extends unknown[] ? number : string
    type ReplacementFunc<T> =
        T extends (infer V)[] ? (items: V[]) => readonly V[] :
        (items: [string, T[keyof T]][]) => readonly [string, T[keyof T]][]
}